#ifndef _SYSCALL_H
#define _SYSCALL_H

#include "_9syscall.h"
#include "_9.h"
#include <syscall.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <errno.h>
#include <stdbool.h>

typedef long long _sysint_t;
extern bool _init_done;

static __inline _sysint_t __syscall6(
        _sysint_t n,
        _sysint_t a1,
        _sysint_t a2,
        _sysint_t a3,
        _sysint_t a4,
        _sysint_t a5,
        _sysint_t a6
    )
{
    _STRACE_PRINT("strace: %lld - %lld %lld %lld %lld %lld %lld\n",
                n, a1, a2, a3, a4, a5, a6);
    int ret;
    _9siglock();
    switch (n)
    {
    case SYS_exit:
    case SYS_exit_group:
        _9exit(a1);

    case SYS_read: ret = _9read(a1, (void *) a2, a3); break;
    case SYS_write: ret = _9write(a1, (const void *) a2, a3); break;
    case SYS_open: ret = _9open((const char *) a1, a2, a3); break;
    case SYS_close: ret = _9close(a1); break;
    case SYS_creat: ret = _9creat((const char *) a1, a2); break;
    case SYS_readv: ret = _9readv(a1, (const struct iovec *) a2, a3); break;
    case SYS_writev: ret = _9writev(a1, (const struct iovec *) a2, a3); break;
    case SYS_ioctl: ret = _9ioctl(a1, a2, a3); break;
    case SYS_pread64: ret = _9pread(a1, (void *) a2, a3, a4); break;
    case SYS_pwrite64: ret = _9pwrite(a1, (const void *) a2, a3, a4); break;
    case SYS_poll: ret = _9poll((struct pollfd *) a1, a2, a3); break;
    case SYS_select: ret = _9select(
                             a1,
                             (fd_set *) a2,
                             (fd_set *) a3,
                             (fd_set *) a4,
                             (struct timeval *restrict) a5); break;
    case SYS_fcntl: ret = _9fcntl(a1, a2, a3); break;
    case SYS_unlink: ret = _9unlink((const char *) a1); break;
    case SYS_readlink: ret = _9readlink((const char *) a1, (char *) a2, a3);

    case SYS_mmap: return (_sysint_t) _9mmap((void *) a1, a2, a3, a4, a5, a6); break;
    case SYS_munmap: ret = _9munmap((void *) a1, a2); break;
    case SYS_brk: return brk((void *) a1); break;

    case SYS_clock_gettime: ret = _9clock_gettime(a1, (struct timespec *) a2); break;

    case SYS_fork: ret = _9fork(); break;

    case SYS_getpid: ret = _9getpid(); break;
    case SYS_getppid: ret = _9getppid(); break;
    case SYS_gettid: ret = _9gettid(); break;

    case SYS_set_thread_area:
        ret = _9set_thread_area((struct user_desc *) a1);
        break;
    case SYS_set_tid_address: ret = _9set_tid_address((int *) a1); break;

    case SYS_arch_prctl: ret = _9arch_prctl(a1, (unsigned long *) a2); break;

    case SYS_rt_sigaction:
        ret = _9rt_sigaction(
                a1,
                (const struct sigaction *)
                a2, (struct sigaction *)
                a3,
                a4);
        break;
    case SYS_rt_sigprocmask:
        ret = _9rt_sigprocmask(a1, (sigset_t *) a3, (sigset_t *) a3, a4);
        break;

    case SYS_stat: ret = _9stat((const char *) a1, (struct stat *) a2); break;
    case SYS_lstat: ret = _9stat((const char *) a1, (struct stat *) a2); break;
    case SYS_fstat: ret = _9fstat(a1, (struct stat *) a2); break;
    case SYS_truncate: ret = _9truncate((const char *) a1, a2); break;
    case SYS_ftruncate: ret = _9ftruncate(a1, a2); break;

    case SYS_getcwd: return (_sysint_t) _9getcwd((char *) a1, a2); break;
    case SYS_getrandom: return _9getrandom((void *) a1, a2, a3);

    default:
        _FAIL_PRINT("cross9: Unimplemented linux syscall "
                "(%lld - %lld %lld %lld %lld %lld %lld)\n",
                n, a1, a2, a3, a4, a5, a6);
        ret = -ENOSYS;
    }

    _9sigunlock();
    return ret;
}

static __inline _sysint_t __syscall5(
        _sysint_t n,
        _sysint_t a1,
        _sysint_t a2,
        _sysint_t a3,
        _sysint_t a4,
        _sysint_t a5
    )
{
    return __syscall6(n, a1, a2, a3, a4, a5, 0);
}

static __inline _sysint_t __syscall4(
        _sysint_t n,
        _sysint_t a1,
        _sysint_t a2,
        _sysint_t a3,
        _sysint_t a4
    )
{
    
    return __syscall6(n, a1, a2, a3, a4, 0, 0);
}

static __inline _sysint_t __syscall3(
        _sysint_t n,
        _sysint_t a1,
        _sysint_t a2,
        _sysint_t a3
    )
{
    
    return __syscall6(n, a1, a2, a3, 0, 0, 0);
}

static __inline _sysint_t __syscall2(
        _sysint_t n,
        _sysint_t a1,
        _sysint_t a2
    )
{
    
    return __syscall6(n, a1, a2, 0, 0, 0, 0);
}

static __inline _sysint_t __syscall1(_sysint_t n, _sysint_t a1)
{
    
    return __syscall6(n, a1, 0, 0, 0, 0, 0);
}

static __inline _sysint_t __syscall0(_sysint_t n)
{
    
    return __syscall6(n, 0, 0, 0, 0, 0, 0);
}

#define SYSCALL_NO_TLS 1
#define IPC_64 0
#endif
