#include <sys/stat.h>
#include <_9syscall.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include "util.h"


static void statconv(const int8_t *restrict buf, struct stat *restrict statbuf)
{
	memset(statbuf, 0, sizeof(*statbuf));

	buf += 2; /* size[2] */
	buf += 2; /* type[2] */
	statbuf->st_dev = le2m(buf, 4);
	buf += 4; /* dev[4] */
	buf += 13; /* qid[1 + 4 + 8] */

	int32_t mode = le2m(buf, 4);
	buf += 4;
	int32_t atime = le2m(buf, 4);
	buf += 4;
	int32_t mtime = le2m(buf, 4);
	buf += 4;

	uint64_t length = le2m(buf, 8);
	buf += 8;

	if (mode & 0x80000000) statbuf->st_mode |= S_IFDIR;
	else if (length == 0) statbuf->st_mode |= S_IFCHR;
	else statbuf->st_mode |= S_IFREG;
	statbuf->st_mode |= mode & 0777;

	statbuf->st_nlink = 1;
	statbuf->st_uid = 1000;
	statbuf->st_gid = 1000;
	statbuf->st_size = length;
	statbuf->st_blksize = 0x1000 > length ? 0x1000 : length;
	statbuf->st_blocks = length  / 512;
	statbuf->st_atime = atime;
	statbuf->st_mtime = mtime;
	statbuf->st_ctime = mtime;
}

int _9stat(const char *restrict pathname, struct stat *restrict statbuf)
{
	int8_t buf[65535];
	if (pathname[0] == 0) return ENOENT;
	int ret = _9syscall_3(SYS_FSTAT, (uvlong) pathname, (uvlong) buf, 65535);
	if (ret == -1) return -EBADF;
	else if (ret == 65535) return -EOVERFLOW;
	statconv(buf, statbuf);
	return 0;
}

int _9fstat(int fd, struct stat *restrict statbuf)
{
	int8_t buf[65535];
	int ret = _9syscall_3(SYS_FSTAT, fd, (uvlong) buf, 65535);
	if (ret == -1) return -EBADF;
	else if (ret == 65535) return -EOVERFLOW;
	statconv(buf, statbuf);
	return 0;
}

int _9ftruncate(int fd, off_t length)
{
	int8_t buf[65536];
	int ret = _9syscall_3(SYS_FSTAT, fd, (uvlong) buf, 65535);
	if (ret == -1) return -EBADF;
	m2le(length, buf + 33, 8); /* 33 is the offset of the length value */
	ret = _9syscall_3(SYS_FWSTAT, fd, (uvlong) buf, 65536);
	if (ret == -1) return -EINVAL;
	return 0;
}

int _9truncate(const char *filename, off_t length)
{
	int8_t buf[65536];
	int ret = _9syscall_3(SYS_STAT, (uvlong) filename, (uvlong) buf, 65535);
	if (ret == -1) return -EBADF;
	m2le(length, buf + 33, 8); /* 33 is the offset of the length value */
	ret = _9syscall_3(SYS_WSTAT, (uvlong) filename, (uvlong) buf, 65536);
	if (ret == -1) return -EINVAL;
	return 0;
}
